""" HTML highlighted code export and language tools """

import os.path
import pygments
from pygments.lexers import get_all_lexers, get_lexer_by_name
from pygments.formatters.html import HtmlFormatter


# ==================================================================== #
#                     Supported langages database                      #
# ==================================================================== #

languages = [
    # (ext, lang)
    ('c', 'c'),
    ('cpp', 'cpp'),
    ('cs', 'csharp'),
    ('css', 'css'),
    ('dart', 'dart'),
    ('diff', 'diff'),
    ('erl', 'erlang'),
    ('ex', 'elixir'),
    ('go', 'go'),
    ('h', 'objectivec'),
    ('hs', 'haskell'),
    ('html', 'html'),
    ('ini', 'ini'),
    ('java', 'java'),
    ('js', 'javascript'),
    ('json', 'json'),
    ('jl', 'julia'),
    ('kt', 'kotlin'),
    ('less', 'less'),
    ('lisp', 'lisp'),
    ('lua', 'lua'),
    ('md', 'markdown'),
    ('php', 'php'),
    ('pl', 'perl'),
    ('py', 'python'),
    ('rb', 'ruby'),
    ('rs', 'rust'),
    ('sass', 'sass'),
    ('scala', 'scala'),
    ('scss', 'scss'),
    ('sh', 'bash'),
    ('sql', 'sql'),
    ('swift', 'swift'),
    ('toml', 'toml'),
    ('ts', 'typescript'),
    ('txt', 'text'),
    ('xml', 'xml'),
    ('yml', 'yaml'),
]
exttolang = {ext: lang for ext, lang in languages}
langtoext = {lang: ext for ext, lang in languages}

def _load_more_languages():
    """ Save more langs to the ``exttolang`` and ``langtoext`` maps """
    for name, aliases, globs, _mimetypes in get_all_lexers():
        if not globs:
           continue
        name = name.lower()
        for glob in globs:
            ext = os.path.splitext(glob)[1][1:]
            langtoext.setdefault(name, ext)
            exttolang.setdefault(ext, name)
        for alias in aliases:
            ext = langtoext[name]
            langtoext.setdefault(alias, ext)
            exttolang.setdefault(ext, alias)

_load_more_languages()
del _load_more_languages

# ==================================================================== #
#                 Plaintext to stylized HTML utilities                 #
# ==================================================================== #

class _TableHtmlFormatter(HtmlFormatter):
    """
    Extension to the default pygment HtmlFormatter to control the html
    skeleton output, class names and line numbering.
    """
    def __init__(self, **options):
        super().__init__(**options)
        if options.get('linenos', False) == 'bin-table':
            self.linenos = 3

    def wrap(self, source, outfile=None):
        if self.linenos == 3:
            source = self._wrap_table(source)
        yield from source

    def _wrap_table(self, inner):
        yield 0, '<table class="highlight"><tbody>'
        for i, (t, l) in enumerate([*inner, (1, '')]):
            yield t, f'<tr><td class="line-number" id=L{i + 1} value={i + 1}></td><td class="line-content">{l}</td></tr>\n'
        yield 0, '</tbody></table>'


_html_formatter = _TableHtmlFormatter(linenos='bin-table', style='monokai')


def highlight(code, language):
    """ Pretty html export of ``code`` using syntax highlighting """
    lexer = get_lexer_by_name(language, startinline=True)
    return pygments.highlight(code, lexer, _html_formatter)



""" Application configuration

The configuration is loaded from the environment variables, you may
override them "manually" or via a ``.env`` file located in the current
directory. Alternativaly you can use the ``--rtdbin-config`` command line
option to define a dotenv to load instead of ``./.env``.

It tries to correctly configure logging according to the server software
you are using

The options are:

* ``RTDBIN_HOST``: The HTTP address the embedded server binds (default: localhost)
* ``RTDBIN_PORT``: The HTTP port the embedded server binds (default: 8012)
* ``RTDBIN_MAXSIZE``: The maximum ``Content-Length`` accepted for new snippets (default: 16kiB)
* ``RTDBIN_IDENTSIZE``: The length of generated unique identifier for new snippets (default: 6)
* ``RTDBIN_DEFAULT_LANGUAGE``: The default language selected in the new snippet form
* ``RTDBIN_DEFAULT_MAXUSAGE``: The default maximum usages before a snippet is automatically removed, 0 means no maximum (default: 0)
* ``RTDBIN_DEFAULT_LIFETIME``: The default time before a snippet is automatically removed, 0 means no limit (default: 0)
* ``REDIS_HOST``: The Redis host adress to connect to
* ``REDIS_PORT``: The Redis port to connect to
* ``REDIS_DB``: The Redis database to connect to
* ``REDIS_PASSWORD``: The password of the Redis database to connect to
* ``REDIS_USERNAME``: The username of the Redis database to connect to
"""


import logging
import os
from argparse import ArgumentParser
from dotenv import load_dotenv
from metrics import Byte, Time
from pprint import pformat


logger = logging.getLogger(__name__)
bin_logger = logging.getLogger(__package__)


def strtobool(s):
    try:
        s = s.lower()
    except AttributeError:
        pass
    return s not in {False, "0", "false", "no"}


def asdict():
    return {
        key: value for key, value in globals().items()
        if key.isupper() and not key.startswith('__')
    }


def _setup():
    global HOST, PORT, MAXSIZE, IDENTSIZE
    global DEFAULT_LANGUAGE, DEFAULT_MAXUSAGE, DEFAULT_LIFETIME
    global REDIS_HOST, REDIS_PORT, REDIS_DB, REDIS_PASSWORD, REDIS_USERNAME

    cli = ArgumentParser()
    cli.add_argument('--rtdbin-port', metavar="PORT", type=int, help="builtin server port")
    cli.add_argument('--rtdbin-config', metavar="PATH", help="dotenv config file")
    options = cli.parse_known_args()[0]
    load_dotenv(options.rtdbin_config)  # use a sensitive default when omitted

    server_software = os.getenv("SERVER_SOFTWARE", "")  # Set by WSGI servers

    # Extact configuration from environ
    if server_software == "":
        HOST = os.getenv('RTDBIN_HOST', 'localhost')
        PORT = options.rtdbin_port or int(os.getenv('RTDBIN_PORT', 8012))
    MAXSIZE = Byte(os.getenv('RTDBIN_MAXSIZE', '16kiB'))
    IDENTSIZE = int(os.getenv('RTDBIN_IDENTSIZE', 6))
    DEFAULT_LANGUAGE = os.getenv('RTDBIN_DEFAULT_LANGUAGE', 'text')
    DEFAULT_MAXUSAGE = int(os.getenv('RTDBIN_DEFAULT_MAXUSAGE', 0))
    DEFAULT_LIFETIME = Time(os.getenv('RTDBIN_DEFAULT_LIFETIME', 0))
    REDIS_HOST = os.getenv('REDIS_HOST', 'localhost')
    REDIS_PORT = int(os.getenv('REDIS_PORT', 6379))
    REDIS_DB = int(os.getenv('REDIS_DB', 0))
    REDIS_PASSWORD = os.getenv('REDIS_PASSWORD') or None
    REDIS_USERNAME = os.getenv('REDIS_USERNAME') or None

    # Configure logging
    if "gunicorn" in server_software:
        gunicorn_logger = logging.getLogger('gunicorn.error')
        bin_logger.handlers = gunicorn_logger.handlers
        bin_logger.setLevel(gunicorn_logger.level)
    elif server_software == "":
        logging.basicConfig(level='DEBUG')
    else:
        logger.warning(
            "Unknown WSGI server: %s, logging may not work as intended.",
            server_software
        )

    # Report config
    logger.debug(
        "Detected WSGI server: %s.",
        server_software or 'builtin wsgiref (unsafe for production)'
    )
    logger.debug("Runtime configuration:\n%s", pformat(asdict()))

_setup()



import bottle

root = __import__("pathlib").Path(__file__).resolve().parent
bottle.TEMPLATE_PATH = [str(root / 'views')]
app = application = bottle.default_app()

from . import config
from . import controller
from . import models



from html.parser import HTMLParser

pairtags = {
    'a', 'abbr', 'acronym', 'address', 'applet', 'article', 'aside', 'audio',
    'b', 'basefont', 'bdi', 'bdo', 'big', 'blockquote', 'body', 'button',
    'canvas', 'caption', 'center', 'cite', 'code', 'colgroup', 'data',
    'datalist', 'dd', 'del', 'details', 'dfn', 'dialog', 'dir', 'div', 'dl',
    'doctype', 'dt', 'em', 'fieldset', 'figcaption', 'figure', 'font', 'footer',
    'form', 'frame', 'frameset', 'head', 'header', 'hgroup', 'h1', 'h2', 'h3',
    'h4', 'h5', 'h6', 'html', 'i', 'iframe', 'ins', 'kbd', 'label', 'legend',
    'li', 'main', 'map', 'mark', 'menu', 'menuitem', 'meter', 'nav', 'noframes',
    'noscript', 'object', 'ol', 'output', 'p', 'picture', 'pre', 'progress',
    'q', 'rp', 'rt', 'ruby', 's', 'samp', 'script', 'section', 'select',
    'small', 'span', 'strike', 'strong', 'style', 'sub', 'summary', 'sup',
    'svg', 'table', 'tbody', 'td', 'template', 'textarea', 'tfoot', 'th',
    'thead', 'time', 'title', 'tr', 'tt', 'u', 'ul', 'var', 'video',
}
singletags = {
    'area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input',
    'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr',
}


class HTMLSanitizer(HTMLParser):

    def __init__(self, testcase):
        self.testcase = testcase
        self.stack = []
        self.insvg = False
        super().__init__()

    def handle_starttag(self, starttag, attrs):
        if self.stack and self.stack[-1] == 'svg':
            return

        if starttag in ('option', 'optgroup'):
            # <option> and <optgroup> can be single or pair, no need to verify
            # append them the the stack, we just verify their immediate parent
            # tag is a <select> or a <datalist>.
            self.testcase.assertTrue(self.stack[-1] in ('select', 'datalist'),
                "%s must be direct child of select or datalist")
            return

        self.testcase.assertIn(starttag, pairtags | singletags, "unknown html tag")
        if starttag in pairtags:
            self.stack.append(starttag)

    def handle_startendtag(self, startendtag, attrs):
        if self.stack and self.stack[-1] == 'svg':
            return

        if startendtag in ('option', 'optgroup'):
            # <option> and <optgroup> can be single or pair, we just verify
            # their immediate parent tag is a <select> or a <datalist>.
            self.testcase.assertTrue(self.stack[-1] in ('select', 'datalist'),
                "%s must be direct child of select or datalist")
            return

        self.testcase.assertIn(startendtag, singletags)

    def handle_endtag(self, endtag):
        if self.stack[-1] == 'svg' and endtag != 'svg':
            return

        if endtag in ('option', 'optgroup'):
            # <option> and <optgroup> can be single or pair, they were not
            # added to the stack, we just verify their immediate parent is
            # a <select> or a <datalist>
            self.testcase.assertTrue(self.stack[-1] in ('select', 'datalist'),
                "%s must be direct child of select or datalist")
            return

        starttag = self.stack.pop()
        self.testcase.assertEqual(starttag, endtag, "malformatted html")



from collections import ChainMap, UserDict
from metrics import Byte, Time
from bin import config

real_config = config.asdict()
default_config = {
    'MAXSIZE': Byte('16kiB'),
    'DEFAULT_LANGUAGE': 'text',
    'DEFAULT_MAXSIZE': 0,
    'DEFAULT_LIFETIME': Time(0),
}

class FakeConfig(UserDict):
    def __init__(self):
        self.data = ChainMap({}, default_config, real_config)

    def __getattr__(self, attr):
        return self[attr]


import unittest
import pygments
from bin import highlight

class TestHighlight(unittest.TestCase):
    @staticmethod
    def has_language_formatter(lang):
        try:
            pygments.lexers.get_lexer_by_name(lang)
            return True
        except pygments.util.ClassNotFound:
            return False

    def test_languages_has_formatter(self):
        for _ext, lang in highlight.languages:
            self.assertTrue(self.has_language_formatter(lang))



""" Mapping between python objets and the Redis store """


from redis import Redis
from genpw import pronounceable_passwd
from bin import config


# We always connect to Redis
database = Redis(
    host=config.REDIS_HOST,
    port=config.REDIS_PORT,
    db=config.REDIS_DB,
    password=config.REDIS_PASSWORD,
    username=config.REDIS_USERNAME,
)


class Snippet:
    """
    A snippet is a immuable text that have been saved in the database
    and that is retrivable via an unique URL.
    """

    def __init__(self, ident, code, views_left, parentid, token):
        self.id = ident  #: snippet unique identifier
        self.code = code  #: snippet text
        self.views_left = views_left  #: how many time this snippet can be retrieved again
        self.parentid = parentid  #: the original snippet this one is a duplicate of or an empty string
        self.token = token #: the admin token of the snippet

    @classmethod
    def new_id(cls):
        """ Generate a safe unique identifier """
        for _ in range(20):
            ident = pronounceable_passwd(config.IDENTSIZE)

            if len(ident) != config.IDENTSIZE:
                continue
            if ident in {'health', 'assets', 'new', 'raw', 'report'}:
                continue
            if database.exists(ident):
                continue

            return ident

        raise RuntimeError("No free or valid identifier has been found after 20 attempts")


    @classmethod
    def create(cls, code, maxusage, lifetime, parentid, token=None):
        """
        Save a snippet in the database and return a snippet object

        :param code: the source code utf-8 encoded
        :param maxusage: how many times this snippet can be retrieve before self-deletion
        :param lifetime: how long the snippet is saved before self-deletion
        :param parentid: the original snippet id this new snippet is a duplicate of, empty string for original snippet
        :param token: the "admin" token of the snippet, ``None`` if the snippet has no "admin" token
        """
        ident = cls.new_id()
        database.hset(ident, b'code', code)
        database.hset(ident, b'views_left', maxusage)
        database.hset(ident, b'parentid', parentid)
        if token:
            database.hset(ident, b'token', token)
        if lifetime > 0:
            database.expire(ident, int(lifetime))
        return cls(ident, code, maxusage, parentid, token)

    @classmethod
    def get_by_id(cls, ident):
        """
        Retrieve a snippet from the database and return a snippet object

        :param ident: the snippet identifier
        :raises KeyError: the snippet does not exist or have been removed
        """
        snippet = database.hgetall(ident)

        if not snippet:
            raise KeyError('Snippet not found')

        code = snippet[b'code'].decode('utf-8')
        views_left = int(snippet[b'views_left'].decode('utf-8'))
        parentid = snippet[b'parentid'].decode('ascii')
        token = snippet.get(b'token', b'').decode() or None
        if views_left == 0:
            pass
        elif views_left == 1:
            database.delete(ident)
        else:
            database.hincrby(ident, 'views_left', -1)

        return cls(ident, code, views_left, parentid, token)

    def delete(self):
        """
        Delete the snippet from the database
        """
        database.delete(self.id)



""" HTTP interface and processing

Various HTTP routes the external world uses to communicate with the application.
"""

import bottle as bt
import cgi
import logging
import re
import os.path
import secrets
from metrics import Time
from bin import root, config, models
from bin.highlight import highlight, languages, langtoext, exttolang


logger = logging.getLogger(__name__)
BOTUARE = re.compile(r'|'.join([
    re.escape('Mozilla/5.0 (compatible; Discordbot/2.0; +https://discordapp.com)'),
    re.escape('facebookexternalhit/1.1 (+http://www.facebook.com/externalhit_uatext.php)'),
]))


@bt.route('/health', method='GET')
def healthcheck():
    """ Get a dummy string 'alive' to ensure the server is responding """
    return "alive"


@bt.route('/', method='GET')
def get_new_form():
    """
    Get the browser-friendly html form to easily post a new snippet

    :param lang: (query) optional lang that is selected in the lang selection instead of the configured default
    :param parentid: (query) optional 'parent' snippet to duplicate the code from

    :raises HTTPError: code 404 when the 'parent' snippet is not found
    """
    parentid = bt.request.query.parentid
    lang = bt.request.query.lang or config.DEFAULT_LANGUAGE

    try:
        code = models.Snippet.get_by_id(parentid).code if parentid else ""
    except KeyError:
        raise bt.HTTPError(404, "Parent snippet not found")

    return bt.template(
        'newform.html',
        languages=languages,
        default_language=lang,
        code=code,
        parentid=parentid,
    )


@bt.route('/assets/<filepath:path>')
def assets(filepath):
    """
    Get a static css/js/media file that is stored in the filesystem.

    This route exists for developers of bin who wish to run the service
    with minimum system requirements. In production, we suggest you use
    a web server to deliver the static content.
    """
    return bt.static_file(filepath, root=root.joinpath('assets'))


@bt.route('/new', method='POST')
def post_new():
    """
    Post a new snippet and redirect the user to the generated unique URL
    for the snippet.

    :param code: (form) required snippet text, can alternativaly be sent as a Multi-Part utf-8 file
    :param lang: (form) optional language
    :param maxusage: (form) optional maximum download of the snippet before it is deleted
    :param lifetime: (form) optional time (defined in seconds) the snippet is keep in the database before it is deleted
    :param parentid: (form) optional snippet id this new snippet is a duplicate of
    :param token: (form) optional the "admin" token allows you to delete your snippet

    :raises HTTPError: code 411 when the ``Content-Length`` http header is missing
    :raises HTTPError: code 413 when the http request is too large (mostly because the snippet is too long)
    :raises HTTPError: code 400 with a sensible status when the form processing fails
    """
    content_length = bt.request.get_header('Content-Length')
    if not content_length:
        raise bt.HTTPError(411, "Content-Length required")
    if int(content_length) > config.MAXSIZE:
        raise bt.HTTPError(413, f"Payload too large, we accept maximum {config.MAXSIZE}")

    files = bt.request.files
    forms = bt.request.forms

    token = None
    code = None
    lang = None
    ext = None
    maxusage = config.DEFAULT_MAXUSAGE
    lifetime = config.DEFAULT_LIFETIME
    parentid = ''

    try:
        # Form extraction
        if files:
            part = next(files.values())
            charset = cgi.parse_header(part.content_type)[1].get('charset', 'utf-8')
            code = part.file.read(config.MAXSIZE).decode(charset)
            ext = os.path.splitext(part.filename)[1][1:] or langtoext[config.DEFAULT_LANGUAGE]
        if forms:
            # WSGI forces latin-1 decoding, this is wrong, we recode it in utf-8
            code = forms.get('code', '').encode('latin-1').decode() or code
            lang = forms.get('lang') or config.DEFAULT_LANGUAGE
            maxusage = int(forms.get('maxusage') or maxusage)
            lifetime = Time(forms.get('lifetime') or lifetime)
            parentid = forms.get('parentid', '')
            token = forms.get('token')

        # Form validation
        if lang:
            ext = langtoext.get(lang)
            if ext is None:
                logger.warning('Unknown lang %r, using %r.', lang, config.DEFAULT_LANGUAGE)
                lang = config.DEFAULT_LANGUAGE
                ext = langtoext[config.DEFAULT_LANGUAGE]

        if ext:
            lang = exttolang.get(ext)
            if lang is None:
                logger.warning('Unknown file extension %r, using %r.', ext, langtoext[config.DEFAULT_LANGUAGE])
                lang = config.DEFAULT_LANGUAGE
                ext = langtoext[config.DEFAULT_LANGUAGE]
        if not code:
            raise ValueError("Code is missing")
        if maxusage < 0:
            raise ValueError("Maximum usage must be positive")
        if lifetime < 0:
            raise ValueError("Lifetime must be positive")
        if parentid:
            try:
                models.Snippet.get_by_id(parentid)
            except KeyError:
                raise ValueError("Parent does not exist")
        if token and len(token) > 22:
            raise ValueError("Token must not exceed 22 chars as of 16 random bytes base64 encoded")
    except ValueError as exc:
        raise bt.HTTPError(400, str(exc))

    snippet = models.Snippet.create(code, maxusage, lifetime, parentid, token)
    logger.info("New %s snippet of %s chars: %s", lang, len(code), snippet.id)
    bt.redirect(f'/{snippet.id}.{ext}')


@bt.route('/<snippetid>', method='GET')
@bt.route('/<snippetid>.<ext>', method='GET')
def get_html(snippetid, ext=None):
    """
    Get a snippet in a beautiful html page

    :param snippetid: (path) required snippet id
    :param ext: (path) optional language file extension, used to determine the highlight backend
    :param token: (query) optional the "admin" token

    :raises HTTPError: code 404 when the snippet is not found
    """
    if BOTUARE.match(bt.request.headers.get('User-Agent', '')):
        return bt.template('blank.html')

    try:
        snippet = models.Snippet.get_by_id(snippetid)
    except KeyError:
        raise bt.HTTPError(404, "Snippet not found")

    lang = langtoext.get(ext, config.DEFAULT_LANGUAGE)
    ext = langtoext[lang]  # always use the prefered extension for that lang
    codehl = highlight(snippet.code, lang)

    return bt.template(
        'highlight.html',
        languages=languages,
        codehl=codehl,
        lang=lang,
        ext=ext,
        snippetid=snippetid,
        parentid=snippet.parentid,
        token=bt.request.query.token,
    )

@bt.route('/<snippetid>', method='DELETE')
@bt.route('/<snippetid>.<ext>', method='DELETE')
def delete_snippet(snippetid, ext=None):
    """
    Delete a snippet

    :param Authorization: (header) required "Authorization: Token <ADMIN_TOKEN>" the "admin" token

    :raises HTTPError: code 400 when the Authorization (token) is missing
    :raises HTTPError: code 404 when the snippet is not found
    :raises HTTPError: code 401 when the token is incorrect
    """
    auth = bt.request.get_header('Authorization', '').split(None, 1)
    if len(auth) != 2 or auth[0] != 'Token':
        raise bt.HTTPError(400, "Token is missing")

    try:
        snippet = models.Snippet.get_by_id(snippetid)
    except KeyError:
        raise bt.HTTPError(404, "Snippet not found")

    if not (snippet.token and secrets.compare_digest(snippet.token, auth[1])):
        raise bt.HTTPError(401, "Unauthorized")

    snippet.delete()
    logger.info("Snippet %s deleted by user", snippetid)


@bt.route('/raw/<snippetid>', method='GET')
@bt.route('/raw/<snippetid>.<ext>', method='GET')
def get_raw(snippetid, ext=None):
    """
    Get a snippet in plain text without code hightlight

    :param snippetid: (path) required snippet id
    :param ext: (path) ignored parameter
    """
    if BOTUARE.match(bt.request.headers.get('User-Agent', '')):
        return bt.template('blank.html')

    try:
        snippet = models.Snippet.get_by_id(snippetid)
    except KeyError:
        raise bt.HTTPError(404, "Snippet not found")

    bt.response.headers['Content-Type'] = 'text/plain'
    return snippet.code

@bt.route('/report', method='POST')
def report():
    """
    Report a problematic snippet to the system administrator.

    :param snippetid: (form) the reported snippet
    :param name: (form) the name of the user reporting the problem

    :raises HTTPError: code 400 when any of the snippetid or the name is missing
    :raises HTTPError: code 404 when the reported snippet is not found
    """
    name = bt.request.forms.get("name", "").encode('latin-1').decode().strip()
    snippetid = bt.request.forms.get("snippetid")
    if not name:
        raise bt.HTTPError(400, "Missing name")
    if not snippetid:
        raise bt.HTTPError(400, "Missing snippetid")

    try:
        models.Snippet.get_by_id(snippetid)
    except KeyError:
        raise bt.HTTPError(404, "Snippet not found")
    logger.warning("The snippet %s got reported by %s", snippetid, name)

    return bt.HTTPResponse("The snippet have been reported.")



import bottle
import unittest
import urllib.request as urlreq

from bottle import template as bottle_template
from html.parser import HTMLParser
from threading import Thread
from unittest.mock import patch, MagicMock
from urllib.parse import urlencode
from urllib.error import HTTPError

import bin as binmod
from bin.models import Snippet
from .fake_config import FakeConfig
from .html_sanitizer import HTMLSanitizer


def make_snippet(ident, code):
    return Snippet(ident=ident, code=code, views_left=float('+inf'), parentid='', token=None)

snippet_lipsum = make_snippet('lipsum', code="Lorem ipsum dolor sit amet")
snippet_python = make_snippet('egg', code='print("Hello world")')
snippet_htmlxss = make_snippet('htmlxss', code='<script>alert("XSS");</script>')
snippet_classified = make_snippet('classified', code='T_xJV3P^FcvYijzH')
snippet_brainfuck = make_snippet('brainfuck', code='++++++++++[>+>+++>+++++++>++++++++++<<<<-]>>>++++.+++++++++.')

UA_HUMAN = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0"
UA_BOT = "Mozilla/5.0 (compatible; Discordbot/2.0; +https://discordapp.com)"


class TestController(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("Starting builtin server in a daemon thread")

        th = Thread(target=bottle.run, daemon=True, kwargs={
            "host": "localhost",
            "port": 8012,
            "debug": True
        })
        th.start()

        # Wait up to 5 seconds for the server to boot
        for i in range(10):
            th.join(.5)
            try:
                urlreq.urlopen("http://localhost:8012/health")
            except ConnectionRefusedError:
                if i == 9:
                    raise
            else:
                break

        cls.config = FakeConfig()
        patch_config = patch.object(binmod.controller, 'config', cls.config)
        patch_config.start()
        cls.addClassCleanup(patch_config.stop)

        mock_bt_tmpl = MagicMock()
        mock_bt_tmpl.side_effect = bottle.template
        patch_bt_tmpl = patch.object(bottle, 'template', mock_bt_tmpl)
        patch_bt_tmpl.start()
        cls.addClassCleanup(patch_bt_tmpl.stop)

    def setUp(self):
        self.html_sanitizer = HTMLSanitizer(self)
        bottle.template.reset_mock()
        self.config.clear()

    # =====

    def test_healthcheck(self):
        with urlreq.urlopen("http://localhost:8012/health") as res:
            self.assertEqual(res.status, 200)

    def test_get_new_form(self):
        with urlreq.urlopen("http://localhost:8012/") as res:
            bottle.template.assert_called_once()
            self.assertEqual(res.status, 200)
            self.html_sanitizer.feed(res.read().decode())

    def test_missing_parentid(self):
        with patch('bin.models.Snippet') as MockSnippet:
            MockSnippet.get_by_id.side_effect = KeyError('Snippet not found')
            with self.assertRaises(HTTPError) as exc:
                urlreq.urlopen("http://localhost:8012/?parentid=foo")

            self.assertEqual(exc.exception.code, 404)

    def test_get_html(self):
        with patch('bin.models.Snippet') as MockSnippet:
            MockSnippet.get_by_id.return_value = snippet_lipsum
            with urlreq.urlopen("http://localhost:8012/lipsum") as res:
                bottle.template.assert_called_once()
                MockSnippet.get_by_id.assert_called_with(snippet_lipsum.id)
                self.assertEqual(res.status, 200)
                self.html_sanitizer.feed(res.read().decode())

            bottle.template.reset_mock()
            MockSnippet.get_by_id.return_value = snippet_python
            with urlreq.urlopen("http://localhost:8012/egg.py") as res:
                bottle.template.assert_called_once()
                MockSnippet.get_by_id.assert_called_with(snippet_python.id)
                self.assertEqual(res.status, 200)
                self.html_sanitizer.feed(res.read().decode())

    def test_get_raw(self):
        with patch('bin.models.Snippet') as MockSnippet:
            MockSnippet.get_by_id.return_value = snippet_lipsum
            with urlreq.urlopen("http://localhost:8012/raw/lipsum") as res:
                bottle.template.assert_not_called()
                MockSnippet.get_by_id.assert_called_with(snippet_lipsum.id)
                self.assertEqual(res.status, 200)
                self.assertEqual(res.read().decode(), snippet_lipsum.code)


            MockSnippet.get_by_id.return_value = snippet_python
            with urlreq.urlopen("http://localhost:8012/raw/egg.py") as res:
                bottle.template.assert_not_called()
                MockSnippet.get_by_id.assert_called_with(snippet_python.id)
                self.assertEqual(res.status, 200)
                self.assertEqual(res.read().decode(), snippet_python.code)

    def test_get_wrong_extension(self):
        with patch('bin.models.Snippet') as MockSnippet, \
             patch('bin.controller.highlight') as MockHighlight:  # from-imported
            MockSnippet.get_by_id.return_value = snippet_lipsum
            MockHighlight.return_value = ''

            with urlreq.urlopen("http://localhost:8012/lipsum.idontexist") as res:
                self.assertEqual(res.status, 200)

            MockSnippet.get_by_id.assert_called_with(snippet_lipsum.id)
            MockHighlight.assert_called_with(snippet_lipsum.code, 'text')

    def test_get_hidden_extension(self):
        with patch('bin.models.Snippet') as MockSnippet, \
             patch('pygments.highlight') as MockHighlight:
            MockSnippet.get_by_id.return_value = snippet_brainfuck
            MockHighlight.return_value = ''

            with urlreq.urlopen("http://localhost:8012/brainfuck.bf") as res:
                self.assertEqual(res.status, 200)

            MockSnippet.get_by_id.assert_called_with(snippet_brainfuck.id)
            MockHighlight.assert_called_once()
            self.assertEqual(MockHighlight.call_args[0][0], snippet_brainfuck.code)
            self.assertIn('BrainfuckLexer', str(MockHighlight.call_args[0][1]))

    def test_against_xss(self):
        testcase = self

        class HTMLParserXSS(HTMLParser):
            def feed(self, data):
                self.found = False
                super().feed(data)
                testcase.assertTrue(self.found)

            def handle_data(self, data):
                if "XSS" in data:
                    self.found = True
                    # ``data`` is unescaped by the calling method already
                    testcase.assertEqual(data.strip(), snippet_htmlxss.code.strip())

        with patch('bin.models.Snippet') as MockSnippet:
            MockSnippet.get_by_id.return_value = snippet_htmlxss
            with urlreq.urlopen("http://localhost:8012/htmlxss") as res:
                self.assertEqual(res.status, 200)
                HTMLParserXSS().feed(res.read().decode())

            with urlreq.urlopen("http://localhost:8012/raw/htmlxss") as res:
                self.assertEqual(res.status, 200)
                self.assertIn('Content-Type', res.headers)
                mimetype = res.headers['Content-Type'].partition(';')[0]
                self.assertEqual(mimetype, 'text/plain')

    def test_discordbot(self):
        with patch('bin.models.Snippet') as MockSnippet:
            MockSnippet.get_by_id.return_value = snippet_classified

            req = urlreq.Request(
                "http://localhost:8012/raw/classified",
                headers={'User-Agent': UA_BOT},
            )
            with urlreq.urlopen(req) as res:
                MockSnippet.get_by_id.assert_not_called()
                self.assertEqual(res.status, 200)

            req = urlreq.Request(
                "http://localhost:8012/raw/classified",
                headers={'User-Agent': UA_HUMAN},
            )
            with urlreq.urlopen(req) as res:
                MockSnippet.get_by_id.assert_called()
                self.assertEqual(res.status, 200)

    def test_new_form(self):
        with patch('bin.models.Snippet') as MockSnippet:
            MockSnippet.create.return_value = snippet_lipsum
            MockSnippet.get_by_id.return_value = snippet_lipsum

            data = urlencode({'code': snippet_lipsum.code}).encode()
            req = urlreq.Request(
                'http://localhost:8012/new',
                data=data,
                headers={
                    'Content-Length': str(len(data))
                }
            )

            with urlreq.urlopen(req) as res:
                MockSnippet.create.assert_called_with(
                    snippet_lipsum.code,
                    self.config.DEFAULT_MAXUSAGE,
                    self.config.DEFAULT_LIFETIME,
                    '',  # parentid
                    None,  # token
                )
                self.assertEqual(res.url, 'http://localhost:8012/lipsum.txt')
